package vexriscv.plugin

import vexriscv.ip._
import vexriscv._
import spinal.core._
import spinal.lib._
import spinal.lib.bus.amba4.axi.Axi4


class DBusDFCPlugin(val config : DataCacheConfig,
                       memoryTranslatorPortConfig : Any = null,
                       dBusCmdMasterPipe : Boolean = false,
                       dBusCmdSlavePipe : Boolean = false,
                       dBusRspSlavePipe : Boolean = false,
                       relaxedMemoryTranslationRegister : Boolean = false,
                       csrInfo : Boolean = false,
                       dfcWayCount : Int = 0,
                       useDFC : Boolean = true,
                       checkDFC : Boolean = false)  extends Plugin[VexRiscv] with DBusAccessService {
  import config._
  
  val omitL1onDFCHit = useDFC && !checkDFC

  assert(isPow2(cacheSize))
  assert(!(memoryTranslatorPortConfig != null && config.cacheSize/config.wayCount > 4096), "When the D$ is used with MMU, each way can't be bigger than a page (4096 bytes)")

  var dBus  : DataCacheMemBus = null
  var mmuBus : MemoryTranslatorBus = null
  var exceptionBus : Flow[ExceptionCause] = null
  var privilegeService : PrivilegeService = null
  var redoBranch : Flow[UInt] = null

  @dontName var dBusAccess : DBusAccess = null
  override def newDBusAccess(): DBusAccess = {
    assert(dBusAccess == null)
    dBusAccess = DBusAccess()
    dBusAccess
  }

  object MEMORY_ENABLE extends Stageable(Bool)
  object MEMORY_MANAGMENT extends Stageable(Bool)
  object MEMORY_WR extends Stageable(Bool)
  object MEMORY_ADDRESS_LOW extends Stageable(UInt(2 bits))
  object MEMORY_LRSC extends Stageable(Bool)
  object MEMORY_AMO extends Stageable(Bool)
  object IS_DBUS_SHARING extends Stageable(Bool())
  object MEMORY_VIRTUAL_ADDRESS extends Stageable(UInt(32 bits))
  object DFC_HIT extends Stageable(Bool)
  object DFC_DATA extends Stageable(Bits(32 bits))
  object MEM_WRITE_DATA extends Stageable(Bits(32 bits))
  object MEM_WRITE_MASK extends Stageable(Bits(4 bits))


  override def setup(pipeline: VexRiscv): Unit = {
    import Riscv._
    import pipeline.config._

    val decoderService = pipeline.service(classOf[DecoderService])

    val stdActions = List[(Stageable[_ <: BaseType],Any)](
      SRC1_CTRL         -> Src1CtrlEnum.RS,
      SRC_USE_SUB_LESS  -> False,
      MEMORY_ENABLE     -> True,
      RS1_USE          -> True,
      IntAluPlugin.ALU_CTRL -> IntAluPlugin.AluCtrlEnum.ADD_SUB
    )

    val loadActions = stdActions ++ List(
      SRC2_CTRL -> Src2CtrlEnum.IMI,
      REGFILE_WRITE_VALID -> True,
      BYPASSABLE_EXECUTE_STAGE -> False,
      BYPASSABLE_MEMORY_STAGE -> False,
      MEMORY_WR -> False
    ) ++ (if(catchSomething) List(HAS_SIDE_EFFECT -> True) else Nil)

    val storeActions = stdActions ++ List(
      SRC2_CTRL -> Src2CtrlEnum.IMS,
      RS2_USE -> True,
      MEMORY_WR -> True
    )

    decoderService.addDefault(MEMORY_ENABLE, False)
    decoderService.add(
      List(LB, LH, LW, LBU, LHU, LWU).map(_ -> loadActions) ++
      List(SB, SH, SW).map(_ -> storeActions)
    )

    def MANAGEMENT  = M"-------00000-----101-----0001111"

    decoderService.addDefault(MEMORY_MANAGMENT, False)
    decoderService.add(MANAGEMENT, List(
      MEMORY_MANAGMENT -> True
    ))

    decoderService.add(FENCE, Nil)

    mmuBus = pipeline.service(classOf[MemoryTranslator]).newTranslationPort(MemoryTranslatorPort.PRIORITY_DATA ,memoryTranslatorPortConfig)
    redoBranch = pipeline.service(classOf[JumpService]).createJumpInterface(if(pipeline.writeBack != null) pipeline.writeBack else pipeline.execute)

    if(catchSomething)
      exceptionBus = pipeline.service(classOf[ExceptionService]).newExceptionPort(if(pipeline.writeBack == null) pipeline.memory else pipeline.writeBack)

    if(pipeline.serviceExist(classOf[PrivilegeService]))
      privilegeService = pipeline.service(classOf[PrivilegeService])

    pipeline.update(DEBUG_BYPASS_CACHE, False)
  }

  override def build(pipeline: VexRiscv): Unit = {
    import pipeline._
    import pipeline.config._

    dBus = master(DataCacheMemBus(this.config)).setName("dBus")

    val cache = new DataCache(this.config.copy(
      mergeExecuteMemory = writeBack == null
    ))

    val DFC = new Area {
      import cache.setup._
    
      val validWays = Reg(Bits(dfcWayCount bits))
        
      val readAddress = Bits(32 bits)
      val readMask = Bits(4 bits)
      
      val memWrite = Bool
      val memWriteAddress = Bits(32 bits)
      val memWriteData = Bits(32 bits)
      val memWriteMask = Bits(4 bits)
      
      val wbWrite = Bool
      val wbWriteAddress = Bits(32 bits)
      val wbWriteData = Bits(32 bits)
      val wbWriteMask = Bits(4 bits)
      
      def matchWordAddrs(a: Bits, b: Bits) = a(31 downto 2) === b(31 downto 2)
      
      val readMatch = Bits(dfcWayCount bits)
      
      val wayToDrop = validWays.andR ?
        readMatch.rotateLeft(1)|
        OHMasking.first(~validWays)
        
      
      val fetchAddress = Bits(32 bits)
      val fetchData = Bits(32 bits)
      val fetchL1Way = Bits(wayCount bits)
      val fetchValid = Bool
      val fetchMatch = Bits(dfcWayCount bits)
      
      val evicted = cache.io.cpu.evicted
      
      val ways = for(i <- 0 until dfcWayCount) yield new Area{
        val virtaddr = Reg(Bits(tagRange.length bits))
        val l1set = Reg(Bits(lineRange.length bits))
        val l1way = Reg(Bits(wayCount bits))
        val dataValid = Reg(Bits(wordPerLine bits))
        val data = Reg(Vec(Bits(32 bits),wordPerLine))

        def matchWay(addr: Bits):Bool = (
          validWays(i) &&
          virtaddr === addr(tagRange) &&
          l1set === addr(lineRange)
        )
        
        def setWord(addr: Bits, data: Bits) = {
          dataValid(addr(wordRange).asUInt) := True
          this.data(addr(wordRange).asUInt) := data
        }
        
        readMatch(i) := matchWay(readAddress)
      
        //Write value fetched from memory into a partial line
        fetchMatch(i) := matchWay(fetchAddress)
        when(fetchValid && fetchMatch(i)) {
            setWord(fetchAddress,fetchData)
        }
        
        val freshWay =  fetchValid && ~fetchMatch.orR && wayToDrop(i) // New address
        
        when(freshWay) {//Write fetched value memory fresh line
          dataValid.clearAll
          setWord(fetchAddress,fetchData)
          validWays(i) := True
          l1way := fetchL1Way
          l1set := fetchAddress(lineRange)
          virtaddr := fetchAddress(tagRange)
        }
        
        
        when(evicted.isValid && evicted.set.asBits === l1set) { //Strictly inclusive
          validWays(i) := False
        }
      }
      
      val flushAll = Bool
      when(flushAll) { validWays.clearAll }


      val cacheReadData = MuxOH(readMatch,ways map (_.data))(readAddress(wordRange).asUInt)
      val dataFoundInCache = MuxOH(readMatch,ways map (_.dataValid))(readAddress(wordRange).asUInt) && readMatch.orR
      val dataFreshlyFetched = fetchValid && matchWordAddrs(readAddress,fetchAddress)
      val readDataValid = dataFoundInCache || dataFreshlyFetched
      
      val readData = Bits(32 bits)
      
      when(dataFreshlyFetched) {
        readData := fetchData
      }.otherwise {
        readData := cacheReadData
      }
      
      for(i <- 0 to 3) {
        val range = i*8+7 downto i*8
        when(readDataValid && wbWrite && matchWordAddrs(readAddress,wbWriteAddress) && wbWriteMask(i)) {
          readData(range) := wbWriteData(range)
        }
        when(readDataValid && memWrite && matchWordAddrs(readAddress,memWriteAddress) && memWriteMask(i)) {
          readData(range) := memWriteData(range)
        }
      }
    }
    
    val DFCctrs = new Area {
      val hits = Reg(UInt(32 bits)) init(0)
      val misses = Reg(UInt(32 bits)) init(0)
      val fill_misses = Reg(UInt(32 bits)) init(0)
      val errors = Reg(UInt(32 bits)) init(0)
    }
    
    pipeline plug DFCctrs
    pipeline plug DFC

    //Interconnect the plugin dBus with the cache dBus with some optional pipelining
    def optionPipe[T](cond : Boolean, on : T)(f : T => T) : T = if(cond) f(on) else on
    def cmdBuf = optionPipe(dBusCmdSlavePipe, cache.io.mem.cmd)(_.s2mPipe())
    dBus.cmd << optionPipe(dBusCmdMasterPipe, cmdBuf)(_.m2sPipe())
    cache.io.mem.rsp << optionPipe(dBusRspSlavePipe,dBus.rsp)(_.m2sPipe())

    pipeline plug new Area{
      //Memory bandwidth counter
      val rspCounter = RegInit(UInt(32 bits)) init(0)
      when(dBus.rsp.valid){
        rspCounter := rspCounter + 1
      }
    }

    decode plug new Area {
      import decode._

      when(mmuBus.busy && arbitration.isValid && input(MEMORY_ENABLE)) {
        arbitration.haltItself := True
      }
    }

    def formatRsp(instr: Bits,ladr: UInt,rsp: Bits):Bits = {
      val rspShifted = Bits(32 bits)
      rspShifted := rsp
      switch(ladr){
        is(1){rspShifted(7 downto 0) := rsp(15 downto 8)}
        is(2){rspShifted(15 downto 0) := rsp(31 downto 16)}
        is(3){rspShifted(7 downto 0) := rsp(31 downto 24)}
      }

      val rspFormated = instr(13 downto 12).mux(
        0 -> B((31 downto 8) -> (rspShifted(7) && !instr(14)),(7 downto 0) -> rspShifted(7 downto 0)),
        1 -> B((31 downto 16) -> (rspShifted(15) && ! instr(14)),(15 downto 0) -> rspShifted(15 downto 0)),
        default -> rspShifted //W
      )
      return rspFormated
    }

    def formatWrite(instr:Bits, data: Bits) = instr(13 downto 12).asUInt.mux(
        U(0)    -> data( 7 downto 0) ## data( 7 downto 0) ## data(7 downto 0) ## data(7 downto 0),
        U(1)    -> data(15 downto 0) ## data(15 downto 0),
        default -> data(31 downto 0)
      )

    val thiz = this
    
    execute plug new Area {
      import execute._
      val size = input(INSTRUCTION)(13 downto 12).asUInt
      
      DFC.readAddress := input(SRC1).asBits
                         //((input(SRC1).asBits(31 downto 12) ##
                         // (input(SRC1).asBits(11 downto 5) | input(SRC2).asBits(11 downto 5))) ##
                         // input(SRC_ADD).asBits(4 downto 0))
                        
      
      val DFC_try = arbitration.isValid && input(MEMORY_ENABLE) && ~input(MEMORY_WR)
      val DFC_hit = DFC.readDataValid && (DFC.readAddress === cache.io.cpu.execute.address.asBits) && DFC_try

      when(arbitration.isValid && input(MEMORY_MANAGMENT)){
        DFCctrs.fill_misses := 0
      }.elsewhen(DFC.readMatch.orR && ~DFC.readDataValid){
        DFCctrs.fill_misses := DFCctrs.fill_misses+1
      }
                      
      insert(DFC_HIT) := DFC_hit
      insert(DFC_DATA) := DFC.readData

      DFC.flushAll := arbitration.isValid && input(MEMORY_MANAGMENT)

      cache.io.cpu.execute.isValid := arbitration.isValid && input(MEMORY_ENABLE) && ~(Bool(omitL1onDFCHit) && DFC_hit)
      cache.io.cpu.execute.address := input(SRC_ADD).asUInt
      cache.io.cpu.execute.args.wr := input(MEMORY_WR)
      cache.io.cpu.execute.args.data := formatWrite(input(INSTRUCTION),input(RS2))
      cache.io.cpu.execute.args.size := size

      cache.io.cpu.flush.valid := arbitration.isValid && input(MEMORY_MANAGMENT)
      arbitration.haltItself setWhen(cache.io.cpu.flush.isStall)

      insert(MEMORY_ADDRESS_LOW) := cache.io.cpu.execute.address(1 downto 0)
      insert(MEM_WRITE_MASK) := size.mux(
        0 -> B"4'b0001", 1 -> B"4'b0011", default -> B"4'b1111"
      )|<<cache.io.cpu.execute.address(1 downto 0)
      insert(MEM_WRITE_DATA) := formatWrite(input(INSTRUCTION),cache.io.cpu.execute.args.data);

      when(cache.io.cpu.redo && arbitration.isValid && input(MEMORY_ENABLE)){
        arbitration.haltItself := True
      }

      if(relaxedMemoryTranslationRegister) insert(MEMORY_VIRTUAL_ADDRESS) := cache.io.cpu.execute.address
    }

    val mmuAndBufferStage = if(writeBack != null) memory else execute
    mmuAndBufferStage plug new Area {
      import mmuAndBufferStage._
      DFC.memWriteAddress := input(SRC_ADD)
      DFC.memWrite := arbitration.isValid && input(MEMORY_WR) && input(MEMORY_ENABLE)
      DFC.memWriteData := input(MEM_WRITE_DATA)
      DFC.memWriteMask := input(MEM_WRITE_MASK)

      cache.io.cpu.memory.isValid := arbitration.isValid && input(MEMORY_ENABLE) && (~(Bool(omitL1onDFCHit) && input(DFC_HIT)))
      cache.io.cpu.memory.isStuck := arbitration.isStuck
      cache.io.cpu.memory.isRemoved := arbitration.removeIt
      cache.io.cpu.memory.address := (if(relaxedMemoryTranslationRegister) input(MEMORY_VIRTUAL_ADDRESS) else if(mmuAndBufferStage == execute) cache.io.cpu.execute.address else U(input(REGFILE_WRITE_DATA)))

      cache.io.cpu.memory.mmuBus <> mmuBus
      cache.io.cpu.memory.mmuBus.rsp.isIoAccess setWhen(pipeline(DEBUG_BYPASS_CACHE) && !cache.io.cpu.memory.isWrite)
      
    }

    val managementStage = stages.last
    managementStage plug new Area{
      import managementStage._
      cache.io.cpu.writeBack.isValid := arbitration.isValid && input(MEMORY_ENABLE) && ~(Bool(omitL1onDFCHit) && input(DFC_HIT))
      cache.io.cpu.writeBack.isStuck := arbitration.isStuck
      cache.io.cpu.writeBack.isUser  := (if(privilegeService != null) privilegeService.isUser() else False)
      cache.io.cpu.writeBack.address := U(input(REGFILE_WRITE_DATA))
      //if(withLrSc) cache.io.cpu.writeBack.clearLrsc := service(classOf[IContextSwitching]).isContextSwitching
      
      val writeMask = B(32 bits,
        (31 downto 24) -> input(MEM_WRITE_MASK)(3),
        (23 downto 16) -> input(MEM_WRITE_MASK)(2),
        (15 downto  8) -> input(MEM_WRITE_MASK)(1),
        ( 7 downto  0) -> input(MEM_WRITE_MASK)(0))
      
      //TODO: Fill?
      DFC.fetchAddress := cache.io.cpu.writeBack.address.asBits
      DFC.fetchData := (input(MEMORY_WR)?
           (cache.io.cpu.writeBack.data & ~writeMask | input(MEM_WRITE_DATA) & writeMask)
         | cache.io.cpu.writeBack.data)
      DFC.fetchL1Way := cache.io.cpu.writeBack.l1Way
      DFC.fetchValid := cache.io.cpu.writeBack.isValid && cache.io.cpu.writeBack.isL1 && arbitration.isFiring
      
      
      DFC.wbWriteAddress := input(SRC_ADD)
      DFC.wbWrite := arbitration.isValid && input(MEMORY_WR) && input(MEMORY_ENABLE)
      DFC.wbWriteData := input(MEM_WRITE_DATA)
      DFC.wbWriteMask := input(MEM_WRITE_MASK)

      redoBranch.valid := False
      redoBranch.payload := input(PC)
      arbitration.flushIt setWhen(redoBranch.valid)
      arbitration.flushNext setWhen(redoBranch.valid)

      if(catchSomething) {
        exceptionBus.valid := False //cache.io.cpu.writeBack.mmuMiss || cache.io.cpu.writeBack.accessError || cache.io.cpu.writeBack.illegalAccess || cache.io.cpu.writeBack.unalignedAccess
        exceptionBus.badAddr := U(input(REGFILE_WRITE_DATA))
        exceptionBus.code.assignDontCare()
      }


      when(arbitration.isValid && input(MEMORY_ENABLE)) {
        if (catchAccessError) when(cache.io.cpu.writeBack.accessError) {
          exceptionBus.valid := True
          exceptionBus.code := (input(MEMORY_WR) ? U(7) | U(5)).resized
        }

        if (catchUnaligned) when(cache.io.cpu.writeBack.unalignedAccess) {
          exceptionBus.valid := True
          exceptionBus.code := (input(MEMORY_WR) ? U(6) | U(4)).resized
        }
        if(catchIllegal) when (cache.io.cpu.writeBack.mmuException) {
          exceptionBus.valid := True
          exceptionBus.code := (input(MEMORY_WR) ? U(15) | U(13)).resized
        }

        when(cache.io.cpu.redo) {
          redoBranch.valid := True
          if(catchSomething) exceptionBus.valid := False
        }
      }

      arbitration.haltItself.setWhen(cache.io.cpu.writeBack.haltIt)

      val rspFormatted = formatRsp(input(INSTRUCTION),input(MEMORY_ADDRESS_LOW),
          (input(DFC_HIT)&&Bool(useDFC))?input(DFC_DATA)|cache.io.cpu.writeBack.data)

      when(arbitration.isValid && input(MEMORY_ENABLE)) {
        output(REGFILE_WRITE_DATA) := rspFormatted
        when(arbitration.isFiring) {
          when(input(DFC_HIT)) {
            DFCctrs.hits := DFCctrs.hits+1
            if(!checkDFC) {
                DFCctrs.errors := DFCctrs.errors+Mux(
                  (input(DFC_DATA)&writeMask) === (cache.io.cpu.writeBack.data&writeMask),
                0,1)
              }
          }.otherwise {
            DFCctrs.misses := DFCctrs.misses+1
          }
        }
      }
      
      when(DFC.flushAll) {
        DFCctrs.hits := 0
        DFCctrs.misses := 0
      }
    }

    //Share access to the dBus (used by self refilled MMU)
    if(dBusAccess != null) pipeline plug new Area{
      dBusAccess.cmd.ready := False
      val forceDatapath = False
      when(dBusAccess.cmd.valid){
        decode.arbitration.haltByOther := True
        val exceptionService = pipeline.service(classOf[ExceptionService])
        when(!stagesFromExecute.map(s => s.arbitration.isValid || exceptionService.isExceptionPending(s)).orR){
          when(!cache.io.cpu.redo) {
            cache.io.cpu.execute.isValid := True
            dBusAccess.cmd.ready := !execute.arbitration.isStuck
          }
          cache.io.cpu.execute.args.wr := dBusAccess.cmd.write
          cache.io.cpu.execute.args.data := dBusAccess.cmd.data
          cache.io.cpu.execute.args.size := dBusAccess.cmd.size
          if(withLrSc) cache.io.cpu.execute.args.isLrsc := False
          if(withAmo) cache.io.cpu.execute.args.isAmo := False
          cache.io.cpu.execute.address := dBusAccess.cmd.address  //Will only be 12 muxes
          forceDatapath := True
        }
      }
      execute.insert(IS_DBUS_SHARING) := dBusAccess.cmd.fire


      mmuBus.cmd.bypassTranslation setWhen(mmuAndBufferStage.input(IS_DBUS_SHARING))
      if(mmuAndBufferStage != execute) (cache.io.cpu.memory.isValid setWhen(mmuAndBufferStage.input(IS_DBUS_SHARING)))
      cache.io.cpu.writeBack.isValid setWhen(managementStage.input(IS_DBUS_SHARING))
      dBusAccess.rsp.valid := managementStage.input(IS_DBUS_SHARING) && !cache.io.cpu.writeBack.isWrite && (cache.io.cpu.redo || !cache.io.cpu.writeBack.haltIt)
      dBusAccess.rsp.data := cache.io.cpu.writeBack.data
      dBusAccess.rsp.error := cache.io.cpu.writeBack.unalignedAccess || cache.io.cpu.writeBack.accessError
      dBusAccess.rsp.redo := cache.io.cpu.redo
      component.addPrePopTask{() =>
        when(forceDatapath){
          execute.output(REGFILE_WRITE_DATA) := dBusAccess.cmd.address.asBits
        }
        if(mmuAndBufferStage != execute) mmuAndBufferStage.input(IS_DBUS_SHARING) init(False)
        managementStage.input(IS_DBUS_SHARING) init(False)
        when(dBusAccess.rsp.valid){
          managementStage.input(IS_DBUS_SHARING).getDrivingReg := False
        }
      }
    }

    if(csrInfo){
      val csr = service(classOf[CsrPlugin])
      csr.r(0xCC0, 0 ->  U(cacheSize/wayCount),  20 ->  U(bytePerLine))
      csr.r(0xCCC, DFCctrs.misses)
      csr.r(0xCCD, DFCctrs.hits)
      csr.r(0xCCE, DFCctrs.errors)
    }
  }
}
